import numpy as np
from collections.abc import Iterable
from typing import Callable
from ..hmanifold import HomogenousSphere
from ..timestepper import EulerLie, RKMK4

_MANIFOLDS = {"hmnsphere": HomogenousSphere}
_METHODS = {"E1": EulerLie, "RKMK4": RKMK4}


class Flow:
    """Object which holds the calculated numerical approximation of an ODE.
    The attributes may be accessed by the regular dot syntax, or by
    ```
    Y, T = flow
    ```
    where flow is an instance of a Flow object.

    The object also supports indexing: `flow[i, j]` is equivalent
    to `flow.Y[i, j]`.

    Attributes
    ----------
    T : array
        List of values t at which the function y(t) is approximated
    Y : array
        Two-dimensional array containing numerical solution.
        Column Y[:, i] corresponds to the solution at T[i].
    """

    def __init__(self, Y, T):
        if not isinstance(Y, np.ndarray):
            raise TypeError("Y must be a numpy array")
        if not isinstance(T, Iterable):
            raise TypeError("T must be array-like")
        self.Y = Y
        self.T = T

    def __iter__(self):
        yield from (self.Y.transpose(), self.T)

    def __getitem__(self, key):
        return self.Y[key]


def solve(
    f: Callable[[float, Iterable], Iterable],
    y,
    t_start,
    t_end,
    h,
    manifold: str,
    method: str,
):
    """Use the specified `method` to compute the numerical solution
    to the ODE defined by `f`. The return flow object will contain a
    parameter `Y` evolving on the given argument for `manifold`.

    To print supported methods and manifolds, use `pylie.manifolds()`
    and `pylie.methods()`.


    Parameters
    ----------
    f : Callable[[float, Iterable], Iterable]
        Function defining the differential equation.
        Must have call signature `f(t, y)`.
    y : Iterable
        Initial value
    t_start : number
        Initial time
    t_end : number
        End time
    h : number
        Step length
    manifold : str
        Manifold on which the ODE evolves. Must be one of
        the manifolds supported by `pylie`. Use `pylie.manifolds()`
        to print a list.
    method : str
        Method to use to solve the ODE. Use `pylie.methods()`
        to print a list of available methods.

    Returns
    -------
    Flow
        Flow object containing the solution, containing attributes Y and T.
        They may be accessed in one of the following ways:
        Either
        ```py
        flow = pylie.solve(*args)
        Y, T = flow
        ```
        or
        ```py
        flow = pylie.solve(*args)
        Y = flow.Y
        T = flow.T
        ```
        The flow object has the following attributes:
        T : array
            List of values t at which the function y(t) is approximated
        Y : array
            Two-dimensional array containing numerical solution.
            Column Y[:, i] corresponds to the solution at T[i].

        It also supports indexing: `flow[i, j]` is equivalent
        to `flow.Y[i, j]`.
    """
    hmanifold = _MANIFOLDS[manifold](y)
    timestepper = _METHODS[method](hmanifold.exp, hmanifold.dexpinv, hmanifold.action)
    t = t_start
    T = [t]
    Y = np.array(y)

    while not np.isclose(t_end - t, 0):
        h = min(h, t_end - t)
        hmanifold.y = timestepper.step(f, t, hmanifold.y, h)
        t = t + h
        Y = np.column_stack((Y, hmanifold.y))
        T.append(t)

    return Flow(Y, np.array(T))



class HomogenousManifold:
    """A homogenous manifold is a manifold acted upon by a Lie group action.

    This is the parent class of all homegenous manifolds.
    """

    def __init__(self, *args):
        self.exp = self.lie_algebra.exp
        self.dexpinv = self.lie_algebra.dexpinv
        self.action = self.lie_group.action

    def dist(self, a, b):
        # TODO
        pass

    def origin(self):
        # TODO
        pass

    def project(self, m):
        # TODO
        pass



import numpy as np
from scipy.linalg import expm


class LieAlgebra:
    def exp(self, y):
        return expm(y)

    def dexpinv(self, u, v, order: int):
        ans = v
        if order >= 2:
            c = self.commutator(u, v)
            ans -= (1 / 2) * c
        if order >= 4:
            c = self.commutator(u, c)
            ans += (1 / 12) * c
        if order >= 6:
            raise NotImplementedError("Not yet implemented for order >= 6")
        return ans

    def commutator(self, a, b):
        if a.ndim == 2 and b.ndim == 2 and a.shape == b.shape:
            # a and b are matrices of similar dimensions
            return a @ b - b @ a
        else:
            raise NotImplementedError


class soLieAlgebra(LieAlgebra):
    def exp(self, y):
        if y.size == 3 and y.ndim == 1:
            # Use the rodrigues formula

            # Check if y is the zero vector
            if np.array_equal(y, np.zeros(3)):
                return np.eye(3)

            alpha = np.linalg.norm(y)
            Y = self.matrix(y)
            return (
                np.eye(3)
                + (np.sin(alpha) / alpha) * Y
                + ((1 - np.cos(alpha)) / alpha ** 2) * Y @ Y
            )
        # Otherwise, use the standard expm formula
        # We are here assuming that y is a matrix
        return super().exp(y)

    def dexpinv(self, u, v, _):
        if u.size == 3 and u.ndim == 1:
            # Use Rodrigues formula

            # Check for the zero vector
            if np.array_equal(u, np.zeros(3)):
                return v

            cot = lambda x: 1 / np.tan(x)  # noqa: E731

            alpha = np.linalg.norm(u)
            U = self.matrix(u)

            # Convert v from the matrix representation to the basis representation
            v = np.array([v[2, 1], v[0, 2], v[1, 0]])

            return (
                np.eye(3)
                - 0.5 * U
                - (2 - alpha * cot(0.5 * alpha)) / (2 * alpha ** 2) * U @ U
            ) @ v
        else:
            return super().dexpinv(u, v)

    def matrix(self, y):
        if y.size != 3:
            raise NotImplementedError("Not yet implemented for n != 3")
        u, v, w = y
        return np.array([[0, -w, v], [w, 0, -u], [-v, u, 0]])
